@@ -3,7 +3,7 @@
  */
 
 /* 
- * Copyright (C) 1991-2003 the Free Software Foundation, Inc.
+ * Copyright (C) 1991-2005 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -20,7 +20,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
 #include "awk.h"
@@ -30,7 +30,7 @@ static reg_syntax_t syn;
 /* make_regexp --- generate compiled regular expressions */
 
 Regexp *
-make_regexp(const char *s, size_t len, int ignorecase)
+make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 {
 	Regexp *rp;
 	const char *rerr;
@@ -39,16 +39,25 @@ make_regexp(const char *s, size_t len, int ignorecase)
 	const char *end = s + len;
 	register char *dest;
 	register int c, c2;
-#ifdef MBS_SUPPORT
-	/* The number of bytes in the current multbyte character.
+	static short first = TRUE;
+	static short no_dfa = FALSE;
+	int has_anchor = FALSE;
+
+	/* The number of bytes in the current multibyte character.
 	   It is 0, when the current character is a singlebyte character.  */
 	size_t is_multibyte = 0;
+#ifdef MBS_SUPPORT
 	mbstate_t mbs;
 
 	if (gawk_mb_cur_max > 1)
 		memset(&mbs, 0, sizeof(mbstate_t)); /* Initialize.  */
 #endif
 
+	if (first) {
+		first = FALSE;
+		no_dfa = (getenv("GAWK_NO_DFA") != NULL);	/* for debugging and testing */
+	}
+
 	/* Handle escaped characters first. */
 
 	/*
@@ -61,7 +70,7 @@ make_regexp(const char *s, size_t len, int ignorecase)
 
 	while (src < end) {
 #ifdef MBS_SUPPORT
-		if (gawk_mb_cur_max > 1 && !is_multibyte) {
+		if (gawk_mb_cur_max > 1 && ! is_multibyte) {
 			/* The previous byte is a singlebyte character, or last byte
 			   of a multibyte character.  We check the next character.  */
 			is_multibyte = mbrlen(src, end - src, &mbs);
@@ -73,12 +82,9 @@ make_regexp(const char *s, size_t len, int ignorecase)
 		}
 #endif
 
-		if (
-#ifdef MBS_SUPPORT
 		/* We skip multibyte character, since it must not be a special
 		   character.  */
-		    (gawk_mb_cur_max == 1 || ! is_multibyte) &&
-#endif
+		if ((gawk_mb_cur_max == 1 || ! is_multibyte) &&
 		    (*src == '\\')) {
 			c = *++src;
 			switch (c) {
@@ -131,12 +137,14 @@ make_regexp(const char *s, size_t len, int ignorecase)
 				src++;
 				break;
 			} /* switch */
-		} else
+		} else {
+			c = *src;
+			if (c == '^' || c == '$')
+				has_anchor = TRUE;
 			*dest++ = *src++;	/* not '\\' */
-#ifdef MBS_SUPPORT
+		}
 		if (gawk_mb_cur_max > 1 && is_multibyte)
 			is_multibyte--;
-#endif
 	} /* while */
 
 	*dest = '\0' ;	/* Only necessary if we print dest ? */
@@ -145,58 +153,107 @@ make_regexp(const char *s, size_t len, int ignorecase)
 	rp->pat.allocated = 0;	/* regex will allocate the buffer */
 	emalloc(rp->pat.fastmap, char *, 256, "make_regexp");
 
-	if (ignorecase)
-		rp->pat.translate = casetable;
-	else
+	/*
+	 * Lo these many years ago, had I known what a P.I.T.A. IGNORECASE
+	 * was going to turn out to be, I wouldn't have bothered with it.
+	 *
+	 * In the case where we have a multibyte character set, we have no
+	 * choice but to use RE_ICASE, since the casetable is for single-byte
+	 * character sets only.
+	 *
+	 * On the other hand, if we do have a single-byte character set,
+	 * using the casetable should give  a performance improvement, since
+	 * it's computed only once, not each time a regex is compiled.  We
+	 * also think it's probably better for portability.  See the
+	 * discussion by the definition of casetable[] in eval.c.
+	 */
+
+	if (ignorecase) {
+		if (gawk_mb_cur_max > 1) {
+			syn |= RE_ICASE;
+			rp->pat.translate = NULL;
+		} else {
+			syn &= ~RE_ICASE;
+			rp->pat.translate = (char *) casetable;
+		}
+	} else {
 		rp->pat.translate = NULL;
+		syn &= ~RE_ICASE;
+	}
+
+	dfasyntax(syn | (ignorecase ? RE_ICASE : 0), ignorecase ? TRUE : FALSE, '\n');
+	re_set_syntax(syn);
+
 	len = dest - temp;
 	if ((rerr = re_compile_pattern(temp, len, &(rp->pat))) != NULL)
 		fatal("%s: /%s/", rerr, temp);	/* rerr already gettextized inside regex routines */
 
 	/* gack. this must be done *after* re_compile_pattern */
 	rp->pat.newline_anchor = FALSE; /* don't get \n in middle of string */
+	if (dfa && ! no_dfa) {
+		dfacomp(temp, len, &(rp->dfareg), TRUE);
+		rp->dfa = TRUE;
+	} else
+		rp->dfa = FALSE;
+	rp->has_anchor = has_anchor;
 
 	free(temp);
 	return rp;
 }
 
-/* research --- do a regexp search */
+/* research --- do a regexp search. use dfa if possible */
 
 int
-research(Regexp *rp, register const char *str, int start,
-	register size_t len, int need_start)
+research(Regexp *rp, register char *str, int start,
+	register size_t len, int flags)
 {
 	const char *ret = str;
+	int try_backref;
+	int need_start;
+	int no_bol;
+	int res;
+
+	need_start = ((flags & RE_NEED_START) != 0);
+	no_bol = ((flags & RE_NO_BOL) != 0);
+
+	if (no_bol)
+		rp->pat.not_bol = 1;
+
+	/*
+	 * Always do dfa search if can; if it fails, then even if
+	 * need_start is true, we won't bother with the regex search.
+	 *
+	 * The dfa matcher doesn't have a no_bol flag, so don't bother
+	 * trying it in that case.
+	 */
+	if (rp->dfa && ! no_bol) {
+		char save;
+		int count = 0;
+ 		/*
+		 * dfa likes to stick a '\n' right after the matched
+		 * text.  So we just save and restore the character.
+		 */
+		save = str[start+len];
+		ret = dfaexec(&(rp->dfareg), str+start, str+start+len, TRUE,
+					&count, &try_backref);
+		str[start+len] = save;
+	}
 
 	if (ret) {
-		/*
-		 * Passing NULL as last arg speeds up search for cases
-		 * where we don't need the start/end info.
-		 */
-		int res = re_search(&(rp->pat), str, start+len,
+		if (need_start || rp->dfa == FALSE || try_backref) {
+			/*
+			 * Passing NULL as last arg speeds up search for cases
+			 * where we don't need the start/end info.
+			 */
+			res = re_search(&(rp->pat), str, start+len,
 				start, len, need_start ? &(rp->regs) : NULL);
-
-		/*
-		 * A return of -2 indicates that a heuristic in
-		 * regex decided it might allocate too much memory
-		 * on the C stack. This doesn't apply to gawk, which
-		 * uses REGEX_MALLOC. This is dealt with by the
-		 * assignment to re_max_failures in resetup().
-		 * Naetheless, we keep this code here as a fallback.
-		 *
-		 * XXX: The above comment is obsolete; the new regex
-		 * doesn't have an re_max_failures variable. But we
-		 * keep the code here just in case.
-		 */
-		if (res == -2) {
-			/* the 10 here is arbitrary */
-			fatal(_("regex match failed, not enough memory to match string \"%.*s%s\""),
-					(int) (len > 10 ? 10 : len), str + start,
-					len > 10 ? "..." : "");
-		}
-		return res;
+		} else
+			res = 1;
 	} else
-		return -1;
+		res = -1;
+
+	rp->pat.not_bol = 0;
+	return res;
 }
 
 /* refree --- free up the dynamic memory used by a compiled regexp */
@@ -217,8 +274,18 @@ refree(Regexp *rp)
 		free(rp->regs.start);
 	if (rp->regs.end)
 		free(rp->regs.end);
+	if (rp->dfa)
+		dfafree(&(rp->dfareg));
 	free(rp);
 }
+ 
+/* dfaerror --- print an error message for the dfa routines */
+
+void
+dfaerror(const char *s)
+{
+	fatal("%s", s);
+}
 
 /* re_update --- recompile a dynamic regexp */
 
@@ -245,6 +312,10 @@ re_update(NODE *t)
 	}
 	if (t->re_reg != NULL)
 		refree(t->re_reg);
+	if (t->re_cnt > 0)
+		t->re_cnt++;
+	if (t->re_cnt > 10)
+		t->re_cnt = 0;
 	if (t->re_text == NULL || (t->re_flags & CASE) != IGNORECASE) {
 		t1 = force_string(tree_eval(t->re_exp));
 		unref(t->re_text);
@@ -252,7 +323,7 @@ re_update(NODE *t)
 		free_temp(t1);
 	}
 	t->re_reg = make_regexp(t->re_text->stptr, t->re_text->stlen,
-				IGNORECASE);
+				IGNORECASE, t->re_cnt);
 	t->re_flags &= ~CASE;
 	t->re_flags |= IGNORECASE;
 	return t->re_reg;
@@ -278,6 +349,24 @@ resetup()
 		syn |= RE_INTERVALS;
 
 	(void) re_set_syntax(syn);
+	dfasyntax(syn, FALSE, '\n');
+}
+
+/* avoid_dfa --- FIXME: temporary kludge function until we have a new dfa.c */
+
+int
+avoid_dfa(NODE *re, char *str, size_t len)
+{
+	char *end;
+
+	if (! re->re_reg->has_anchor)
+		return FALSE;
+
+	for (end = str + len; str < end; str++)
+		if (*str == '\n')
+			return TRUE;
+
+	return FALSE;
 }
 
 /* reisstring --- return TRUE if the RE match is a simple string match */
@@ -300,7 +389,7 @@ reisstring(const char *text, size_t len, Regexp *re, const char *buf)
 	/* make accessable to gdb */
 	matched = &buf[RESTART(re, buf)];
 
-	res = STREQN(text, matched, len);
+	res = (memcmp(text, matched, len) == 0);
 
 	return res;
 }
@@ -348,6 +437,9 @@ reflags2str(int flagval)
 		{ RE_DEBUG, "RE_DEBUG" },
 		{ RE_INVALID_INTERVAL_ORD, "RE_INVALID_INTERVAL_ORD" },
 		{ RE_ICASE, "RE_ICASE" },
+		{ RE_CARET_ANCHORS_HERE, "RE_CARET_ANCHORS_HERE" },
+		{ RE_CONTEXT_INVALID_DUP, "RE_CONTEXT_INVALID_DUP" },
+		{ RE_NO_SUB, "RE_NO_SUB" },
 		{ 0,	NULL },
 	};
 
