@@ -2,4 +2,4 @@
 AWK=${AWK-../gawk}
 echo $$ > _pid.in
 echo $1 >> _pid.in
-exec $AWK -f pid.awk _pid.in
+exec $AWK -f pid.awk _pid.in 2>/dev/null
